package net.knarcraft.blacksmith.config.blacksmith;

import net.citizensnpcs.api.util.DataKey;
import net.citizensnpcs.api.util.YamlStorage;
import net.knarcraft.blacksmith.BlacksmithPlugin;
import net.knarcraft.blacksmith.config.SettingValueType;
import net.knarcraft.blacksmith.util.ConfigHelper;
import net.knarcraft.blacksmith.util.InputParsingHelper;
import net.knarcraft.blacksmith.util.ItemHelper;
import org.bukkit.Material;
import org.bukkit.enchantments.Enchantment;

import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;

/**
 * A class which keeps track of all default NPC settings and all global settings
 */
public class GlobalBlacksmithSettings {

    private final Map<Material, Double> materialBasePrices = new HashMap<>();
    private final Map<Material, Double> materialPricePerDurabilityPoints = new HashMap<>();
    private final Map<Enchantment, Double> enchantmentCosts = new HashMap<>();
    private final Map<BlacksmithNPCSetting, Object> defaultNPCSettings = new HashMap<>();
    private final List<Material> defaultReforgeAbleMaterials = new ArrayList<>();
    private final List<Enchantment> defaultEnchantmentBlocklist = new ArrayList<>();
    private final Map<GlobalBlacksmithSetting, Object> globalSettings = new HashMap<>();

    private final YamlStorage defaultConfig;

    /**
     * Instantiates a new "Settings"
     *
     * @param plugin <p>A reference to the blacksmith plugin</p>
     */
    public GlobalBlacksmithSettings(BlacksmithPlugin plugin) {
        defaultConfig = new YamlStorage(new File(plugin.getDataFolder() + File.separator + "config.yml"),
                "Blacksmith Configuration\nWarning: The values under defaults are the values set for a " +
                        "blacksmith upon creation. To change any values for existing NPCs, edit the citizens NPC file.");
    }

    /**
     * Loads all configuration values from the config file
     */
    public void load() {
        //Load the config from disk
        defaultConfig.load();
        DataKey root = defaultConfig.getKey("");

        //Just in case, clear existing values
        defaultNPCSettings.clear();
        globalSettings.clear();
        materialBasePrices.clear();
        materialPricePerDurabilityPoints.clear();
        enchantmentCosts.clear();

        //Load/Save NPC default settings
        loadDefaultNPCSettings(root);

        //Load/Save global settings
        loadGlobalSettings(root);

        //Save any modified values to disk
        defaultConfig.save();
    }

    /**
     * Changes the value of the given setting
     *
     * @param globalBlacksmithSetting <p>The global setting to change</p>
     * @param newValue      <p>The new value of the setting</p>
     */
    public void changeValue(GlobalBlacksmithSetting globalBlacksmithSetting, Object newValue) {
        globalSettings.put(globalBlacksmithSetting, newValue);
        save();
    }

    /**
     * Changes the value of the given setting
     *
     * @param blacksmithNpcSetting <p>The default NPC setting to change</p>
     * @param newValue   <p>The new value for the setting</p>
     */
    public void changeValue(BlacksmithNPCSetting blacksmithNpcSetting, Object newValue) {
        if (blacksmithNpcSetting.getValueType() == SettingValueType.STRING_LIST ||
                blacksmithNpcSetting.getValueType() == SettingValueType.REFORGE_ABLE_ITEMS) {
            //Workaround to make sure it's treated as the correct type
            defaultNPCSettings.put(blacksmithNpcSetting, newValue == null ? null : ConfigHelper.asStringList(newValue));
        } else {
            defaultNPCSettings.put(blacksmithNpcSetting, newValue);
        }
        save();
        if (blacksmithNpcSetting == BlacksmithNPCSetting.REFORGE_ABLE_ITEMS) {
            loadReforgeAbleItems();
        } else if (blacksmithNpcSetting == BlacksmithNPCSetting.ENCHANTMENT_BLOCKLIST) {
            loadEnchantmentBlocklist();
        }
    }

    /**
     * Gets the current raw value of the given global setting
     *
     * @param globalBlacksmithSetting <p>The setting to get</p>
     * @return <p>The current raw setting value</p>
     */
    public Object getRawValue(GlobalBlacksmithSetting globalBlacksmithSetting) {
        return globalSettings.get(globalBlacksmithSetting);
    }

    /**
     * Gets the current raw value of the given default NPC setting
     *
     * @param blacksmithNpcSetting <p>The setting to get</p>
     * @return <p>The current raw setting value</p>
     */
    public Object getRawValue(BlacksmithNPCSetting blacksmithNpcSetting) {
        return defaultNPCSettings.get(blacksmithNpcSetting);
    }

    /**
     * Sets the enchantment cost for the given enchantment
     *
     * @param enchantment        <p>The enchantment to set the enchantment cost for</p>
     * @param newEnchantmentCost <p>The new enchantment cost</p>
     */
    public void setEnchantmentCost(Enchantment enchantment, double newEnchantmentCost) {
        if (enchantment == null) {
            if (newEnchantmentCost < 0) {
                throw new IllegalArgumentException("Enchantment cost cannot be negative!");
            }
            globalSettings.put(GlobalBlacksmithSetting.ENCHANTMENT_COST, newEnchantmentCost);
        } else {
            if (newEnchantmentCost < 0) {
                enchantmentCosts.put(enchantment, null);
            } else {
                enchantmentCosts.put(enchantment, newEnchantmentCost);
            }
        }
        save();
    }

    /**
     * Sets the price per durability point for the given material
     *
     * @param material <p>The material to set the price per durability point price for</p>
     * @param newPrice <p>The new price per durability point price</p>
     */
    public void setPricePerDurabilityPoint(Material material, double newPrice) {
        if (material == null) {
            if (newPrice < 0) {
                throw new IllegalArgumentException("Price per durability point cannot be negative!");
            }
            globalSettings.put(GlobalBlacksmithSetting.PRICE_PER_DURABILITY_POINT, newPrice);
        } else {
            //Use a negative price to unset the per-item value
            if (newPrice < 0) {
                materialPricePerDurabilityPoints.put(material, null);
            } else {
                materialPricePerDurabilityPoints.put(material, newPrice);
            }
        }
        save();
    }

    /**
     * Sets the base price for the given material
     *
     * @param material     <p>The material to set the base price for</p>
     * @param newBasePrice <p>The new base price</p>
     */
    public void setBasePrice(Material material, double newBasePrice) {
        if (material == null) {
            if (newBasePrice < 0) {
                throw new IllegalArgumentException("Base price cannot be negative!");
            }
            globalSettings.put(GlobalBlacksmithSetting.BASE_PRICE, newBasePrice);
        } else {
            //Use a negative price to unset the per-item value
            if (newBasePrice < 0) {
                materialBasePrices.put(material, null);
            } else {
                materialBasePrices.put(material, newBasePrice);
            }
        }
        save();
    }

    /**
     * Gets the current value of the default NPC settings
     *
     * @return <p>The current value of the default NPC settings</p>
     */
    public Map<BlacksmithNPCSetting, Object> getDefaultNPCSettings() {
        return new HashMap<>(this.defaultNPCSettings);
    }

    /**
     * Gets whether to use natural cost for cost calculation
     *
     * <p>Natural cost makes it more costly the more damage is dealt to an item. The alternative is the legacy behavior
     * where the amount of durability points remaining increases the cost.</p>
     *
     * @return <p>Whether to use natural cost</p>
     */
    public boolean getUseNaturalCost() {
        return asBoolean(GlobalBlacksmithSetting.NATURAL_COST);
    }

    /**
     * Gets whether to show exact time for reforging wait-time, and for wait-time between sessions
     *
     * @return <p>Whether to show exact time</p>
     */
    public boolean getShowExactTime() {
        return asBoolean(GlobalBlacksmithSetting.SHOW_EXACT_TIME);
    }

    /**
     * Gets the base price for the given material
     *
     * @param material <p>The material to get the base price for</p>
     * @return <p>The base price for the material</p>
     */
    public double getBasePrice(Material material) {
        if (materialBasePrices.containsKey(material) && materialBasePrices.get(material) != null) {
            return materialBasePrices.get(material);
        } else {
            return asDouble(GlobalBlacksmithSetting.BASE_PRICE);
        }
    }

    /**
     * Gets the price per durability point for the given material
     *
     * @param material <p>The material to get the durability point price for</p>
     * @return <p>The durability point price for the material</p>
     */
    public double getPricePerDurabilityPoint(Material material) {
        if (materialPricePerDurabilityPoints.containsKey(material) &&
                materialPricePerDurabilityPoints.get(material) != null) {
            return materialPricePerDurabilityPoints.get(material);
        } else {
            return asDouble(GlobalBlacksmithSetting.PRICE_PER_DURABILITY_POINT);
        }
    }

    /**
     * Gets the cost to be added for each level of the given enchantment
     *
     * @param enchantment <p>The enchantment to get the cost for</p>
     * @return <p>The cost of each enchantment level</p>
     */
    public double getEnchantmentCost(Enchantment enchantment) {
        if (enchantmentCosts.containsKey(enchantment) && enchantmentCosts.get(enchantment) != null) {
            return enchantmentCosts.get(enchantment);
        } else {
            return asDouble(GlobalBlacksmithSetting.ENCHANTMENT_COST);
        }
    }

    /**
     * Gets the value of reforgeAbleItems
     *
     * @return <p>The value of reforgeAbleItems</p>
     */
    public List<Material> getReforgeAbleItems() {
        return this.defaultReforgeAbleMaterials;
    }

    /**
     * Gets the value of enchantmentBlocklist
     *
     * @return <p>The value of enchantmentBlocklist</p>
     */
    public List<Enchantment> getEnchantmentBlocklist() {
        return this.defaultEnchantmentBlocklist;
    }

    /**
     * Gets the cost for repairing the given type of anvil
     *
     * @param material <p>The anvil material to repair</p>
     * @return <p>The cost of repairing the anvil</p>
     */
    public double getAnvilCost(Material material) {
        if (material == Material.CHIPPED_ANVIL) {
            return asDouble(GlobalBlacksmithSetting.ANVIL_CHIPPED_COST);
        } else if (material == Material.DAMAGED_ANVIL) {
            return asDouble(GlobalBlacksmithSetting.ANVIL_DAMAGED_COST);
        } else {
            throw new IllegalArgumentException("An unexpected item was encountered!");
        }
    }

    /**
     * Gets the given value as a boolean
     *
     * <p>This will throw an exception if used for a non-boolean value</p>
     *
     * @param setting <p>The setting to get the value of</p>
     * @return <p>The value of the given setting as a boolean</p>
     */
    public boolean asBoolean(GlobalBlacksmithSetting setting) {
        return ConfigHelper.asBoolean(getValue(setting));
    }

    /**
     * Gets the given value as a double
     *
     * <p>This will throw an exception if used for a non-double setting</p>
     *
     * @param setting <p>The setting to get the value of</p>
     * @return <p>The value of the given setting as a double</p>
     */
    public double asDouble(GlobalBlacksmithSetting setting) {
        return ConfigHelper.asDouble(getValue(setting));
    }

    /**
     * Gets the value of a setting, using the default if not set
     *
     * @param setting <p>The setting to get the value of</p>
     * @return <p>The current value</p>
     */
    private Object getValue(GlobalBlacksmithSetting setting) {
        Object value = globalSettings.get(setting);
        //If not set in config.yml, use the default value from the enum
        if (value == null) {
            value = setting.getDefaultValue();
        }
        return value;
    }

    /**
     * Loads all global settings
     *
     * @param root <p>The root node of all global settings</p>
     */
    private void loadGlobalSettings(DataKey root) {
        for (GlobalBlacksmithSetting globalBlacksmithSetting : GlobalBlacksmithSetting.values()) {
            if (!root.keyExists(globalBlacksmithSetting.getPath())) {
                //If the setting does not exist in the config file, add it
                root.setRaw(globalBlacksmithSetting.getPath(), globalBlacksmithSetting.getDefaultValue());
            } else {
                //Set the setting to the value found in the path
                globalSettings.put(globalBlacksmithSetting, root.getRaw(globalBlacksmithSetting.getPath()));
            }
        }

        //Load all base prices
        loadBasePrices(root);

        //Load all per-durability-point prices
        loadPricesPerDurabilityPoint(root);

        //Load all enchantment prices
        DataKey enchantmentCostNode = root.getRelative(GlobalBlacksmithSetting.ENCHANTMENT_COST.getParent());
        Map<String, String> relevantKeys = getRelevantKeys(enchantmentCostNode);
        for (String key : relevantKeys.keySet()) {
            String enchantmentName = relevantKeys.get(key);
            Enchantment enchantment = InputParsingHelper.matchEnchantment(enchantmentName);
            setItemPrice(enchantmentCosts, enchantmentName, enchantment, enchantmentCostNode.getDouble(key));
        }
    }

    /**
     * Loads all prices per durability point for all materials
     *
     * @param root <p>The configuration root node to search from</p>
     */
    private void loadPricesPerDurabilityPoint(DataKey root) {
        DataKey basePerDurabilityPriceNode = root.getRelative(GlobalBlacksmithSetting.PRICE_PER_DURABILITY_POINT.getParent());
        Map<String, String> relevantKeys = getRelevantKeys(basePerDurabilityPriceNode);

        for (String key : relevantKeys.keySet()) {
            String materialName = relevantKeys.get(key);
            double price = basePerDurabilityPriceNode.getDouble(key);
            if (materialName.contains("*")) {
                //Treat *CHESTPLATE as a regular expression to match all chest-plates
                setMatchedMaterialPrices(materialPricePerDurabilityPoints, materialName, price);
            } else {
                Material material = InputParsingHelper.matchMaterial(materialName);
                setItemPrice(materialPricePerDurabilityPoints, materialName, material, price);
            }
        }
    }

    /**
     * Loads base prices for all materials
     *
     * @param root <p>The configuration root node to search from</p>
     */
    private void loadBasePrices(DataKey root) {
        DataKey basePriceNode = root.getRelative(GlobalBlacksmithSetting.BASE_PRICE.getParent());
        Map<String, String> relevantKeys = getRelevantKeys(basePriceNode);

        for (String key : relevantKeys.keySet()) {
            String materialName = relevantKeys.get(key);
            double price = basePriceNode.getDouble(key);
            if (materialName.contains("*")) {
                //Treat *CHESTPLATE as a regular expression to match all chest-plates
                setMatchedMaterialPrices(materialBasePrices, materialName, price);
            } else {
                Material material = InputParsingHelper.matchMaterial(materialName);
                setItemPrice(materialBasePrices, materialName, material, price);
            }
        }
    }

    /**
     * Sets the price for any materials matching the given wildcard material name
     *
     * @param prices       <p>The map to store the prices in</p>
     * @param materialName <p>The material name to match</p>
     * @param price        <p>The price to set for the matched materials</p>
     */
    private void setMatchedMaterialPrices(Map<Material, Double> prices, String materialName, double price) {
        String search = InputParsingHelper.regExIfy(materialName);
        for (Material material : ItemHelper.getAllReforgeAbleMaterials()) {
            if (material.name().matches(search)) {
                setItemPrice(prices, material.name(), material, price);
            }
        }
    }

    /**
     * Sets the price for the given material
     *
     * @param prices   <p>The map to store the price in</p>
     * @param itemName <p>The name of the material to add a price for</p>
     * @param item     <p>The material parsed from the name</p>
     * @param price    <p>The price to set</p>
     */
    private <K> void setItemPrice(Map<K, Double> prices, String itemName, K item, double price) {
        if (item != null) {
            prices.put(item, price);
        } else {
            BlacksmithPlugin.getInstance().getLogger().log(Level.WARNING,
                    "Unable to find a material/enchantment matching " + itemName);
        }
    }

    /**
     * Gets a map between relevant keys and their normalized name
     *
     * @param rootKey <p>The root data key containing sub-keys</p>
     * @return <p>Any sub-keys found that aren't the default</p>
     */
    private Map<String, String> getRelevantKeys(DataKey rootKey) {
        Map<String, String> relevant = new HashMap<>();
        for (DataKey dataKey : rootKey.getSubKeys()) {
            String keyName = dataKey.name();
            //Skip the default value
            if (keyName.equals("default")) {
                continue;
            }
            String normalizedName = keyName.toUpperCase().replace("-", "_");
            relevant.put(keyName, normalizedName);
        }
        return relevant;
    }

    /**
     * Converts a normalized material name to the format used in the config file
     *
     * @param normalizedName <p>The normalized name to un-normalize</p>
     * @return <p>The un-normalized name</p>
     */
    private String unNormalizeName(String normalizedName) {
        return normalizedName.toLowerCase().replace("_", "-");
    }

    /**
     * Loads all default NPC settings
     *
     * @param root <p>The root node of all default NPC settings</p>
     */
    private void loadDefaultNPCSettings(DataKey root) {
        for (BlacksmithNPCSetting setting : BlacksmithNPCSetting.values()) {
            if (!root.keyExists(setting.getPath())) {
                //If the setting does not exist in the config file, add it
                root.setRaw(setting.getPath(), setting.getDefaultValue());
            } else {
                //Set the setting to the value found in the path
                defaultNPCSettings.put(setting, root.getRaw(setting.getPath()));
            }
        }
        loadReforgeAbleItems();
        loadEnchantmentBlocklist();
    }

    /**
     * Loads reforgeAble items from the current value
     */
    private void loadReforgeAbleItems() {
        defaultReforgeAbleMaterials.clear();
        defaultReforgeAbleMaterials.addAll(BlacksmithNPCSettings.getReforgeAbleItems(ConfigHelper.asStringList(
                defaultNPCSettings.get(BlacksmithNPCSetting.REFORGE_ABLE_ITEMS))));
    }

    /**
     * Loads the enchantment blocklist from the current value
     */
    private void loadEnchantmentBlocklist() {
        defaultEnchantmentBlocklist.clear();
        defaultEnchantmentBlocklist.addAll(BlacksmithNPCSettings.getEnchantmentBlocklist(ConfigHelper.asStringList(
                defaultNPCSettings.get(BlacksmithNPCSetting.ENCHANTMENT_BLOCKLIST))));
    }

    /**
     * Saves all current settings to the config file
     */
    private void save() {
        DataKey root = defaultConfig.getKey("");
        //Save all default NPC settings
        for (BlacksmithNPCSetting setting : BlacksmithNPCSetting.values()) {
            root.setRaw(setting.getPath(), defaultNPCSettings.get(setting));
        }

        //Save all normal global settings
        for (GlobalBlacksmithSetting globalBlacksmithSetting : GlobalBlacksmithSetting.values()) {
            root.setRaw(globalBlacksmithSetting.getPath(), globalSettings.get(globalBlacksmithSetting));
        }

        //Save all base prices
        DataKey basePriceNode = root.getRelative(GlobalBlacksmithSetting.BASE_PRICE.getParent());
        for (Material material : materialBasePrices.keySet()) {
            basePriceNode.setRaw(unNormalizeName(material.name()), materialBasePrices.get(material));
        }

        //Save all per-durability-point prices
        DataKey basePerDurabilityPriceNode = root.getRelative(GlobalBlacksmithSetting.PRICE_PER_DURABILITY_POINT.getParent());
        for (Material material : materialPricePerDurabilityPoints.keySet()) {
            basePerDurabilityPriceNode.setRaw(unNormalizeName(material.name()), materialPricePerDurabilityPoints.get(material));
        }

        //Load all enchantment prices
        DataKey enchantmentCostNode = root.getRelative(GlobalBlacksmithSetting.ENCHANTMENT_COST.getParent());
        for (Enchantment enchantment : enchantmentCosts.keySet()) {
            enchantmentCostNode.setRaw(unNormalizeName(enchantment.getKey().getKey()), enchantmentCosts.get(enchantment));
        }

        //Perform the actual save to disk
        defaultConfig.save();
    }

}
